@@ -31,8 +31,6 @@
 # with Ubiquity; if not, write to the Free Software Foundation, Inc., 51
 # Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 
-from __future__ import print_function
-
 import atexit
 import configparser
 from functools import reduce
@@ -150,8 +148,8 @@ class Controller(ubiquity.frontend.base.Controller):
             self._wizard.navigation_control.hide()
         self._wizard.refresh()
 
-    def toggle_next_button(self, label='gtk-go-forward'):
-        self._wizard.toggle_next_button(label)
+    def toggle_next_button(self, label='gtk-go-forward', suggested=False):
+        self._wizard.toggle_next_button(label, suggested=suggested)
 
     def toggle_skip_button(self, label='skip'):
         self._wizard.toggle_skip_button(label)
@@ -162,9 +160,16 @@ class Controller(ubiquity.frontend.base.Controller):
 
 def on_screen_reader_enabled_changed(gsettings, key):
     # handle starting orca only, it exits itself when the key is false
-    if (key == "screen-reader-enabled" and gsettings.get_boolean(key) and
-       osextras.find_on_path('orca')):
-        subprocess.Popen(['orca'], preexec_fn=misc.drop_all_privileges)
+    if key == "screen-reader-enabled":
+        # Besides starting orca, also make sure the screen-reader-enabled
+        # setting gets passed to the target system.
+        if (gsettings.get_boolean(key) and osextras.find_on_path('orca')):
+            # Enable
+            subprocess.Popen(['orca'], preexec_fn=misc.drop_all_privileges)
+            os.environ['UBIQUITY_A11Y_PROFILE'] = 'screen-reader'
+        elif 'UBIQUITY_A11Y_PROFILE' in os.environ:
+            # Disable
+            del os.environ['UBIQUITY_A11Y_PROFILE']
 
 
 class Wizard(BaseFrontend):
@@ -220,7 +225,9 @@ class Wizard(BaseFrontend):
         self.language_questions = ('live_installer', 'quit', 'back', 'next',
                                    'warning_dialog', 'warning_dialog_label',
                                    'cancelbutton', 'exitbutton',
-                                   'install_button', 'restart_to_continue')
+                                   'install_button',
+                                   'restart_button',
+                                   'restart_to_continue')
         self.current_page = None
         self.backup = None
         self.allowed_change_step = True
@@ -233,6 +240,7 @@ class Wizard(BaseFrontend):
         self.installing_no_return = False
         self.partitioned = False
         self.timezone_set = False
+        self.ubuntu_drivers = None
         self.returncode = 0
         self.history = []
         self.builder = Gtk.Builder()
@@ -245,6 +253,7 @@ class Wizard(BaseFrontend):
         self.screen_reader = False
         self.orca_process = None
         self.a11y_settings = None
+        self.have_apt_updated = False
 
         # To get a "busy mouse":
         self.watch = Gdk.Cursor.new(Gdk.CursorType.WATCH)
@@ -391,11 +400,71 @@ class Wizard(BaseFrontend):
             self.live_installer.connect(
                 'key-press-event', self.a11y_profile_keys)
 
+        # Wait for sound.target to become ready
+        self.soundWatcher = misc.SystemdUnitWatcher('sound.target',
+                                                    self.play_system_ready)
+
+        self.save_oem_metapackages_list()
+
+    def play_system_ready(self):
+        # play the system ready sound
         if osextras.find_on_path('canberra-gtk-play'):
             subprocess.Popen(
                 ['canberra-gtk-play', '--id=system-ready'],
                 preexec_fn=misc.drop_all_privileges)
 
+    def save_oem_metapackages_list(self, wait_finished=False):
+        ''' If we can, update the apt indexes. Then run 'ubuntu-drivers
+        list-oem' to find any OEM metapackages for this system. If we've
+        already done this with updated apt indexes before, there's no point
+        running again, so just return. '''
+
+        # We already did it
+        if self.have_apt_updated:
+            syslog.syslog(syslog.LOG_INFO, "We've already apt updated and "
+                          "run, not doing so again.")
+            return
+
+        with misc.raised_privileges():
+            try:
+                import apt
+                syslog.syslog(syslog.LOG_INFO, "Trying to update apt indexes "
+                              "to run ubuntu-drivers against fresh data.")
+                cache = apt.cache.Cache()
+                cache.update()
+                cache.open()
+                self.have_apt_updated = True
+            except apt.cache.FetchFailedException:
+                syslog.syslog(syslog.LOG_INFO,
+                              "Failed to update apt indexes; offline? "
+                              "Continuing without.")
+            except apt.cache.LockFailedException:
+                syslog.syslog(syslog.LOG_WARNING,
+                              "Failed to update apt indexes, permission "
+                              "denied: running under test?")
+
+            if osextras.find_on_path('ubuntu-drivers'):
+                # We already ran, were offline then, and are still offline, no
+                # point running again.
+                if os.path.exists('/run/ubuntu-drivers-oem.autoinstall') and \
+                        not self.have_apt_updated:
+                    return
+
+                # It's either the first run or a subsequent one but we now
+                # managed to update the apt indexes.
+                self.ubuntu_drivers = subprocess.Popen(
+                    ['ubuntu-drivers',
+                     'list-oem',
+                     '--package-list',
+                     '/run/ubuntu-drivers-oem.autoinstall'],
+                    stdin=subprocess.DEVNULL,
+                    stdout=subprocess.PIPE,
+                    stderr=subprocess.DEVNULL)
+
+                if wait_finished:
+                    self.ubuntu_drivers.communicate()
+                    self.ubuntu_drivers = None
+
     def all_children(self, parent):
         if isinstance(parent, Gtk.Container):
             def recurse(x, y):
@@ -587,6 +656,29 @@ class Wizard(BaseFrontend):
 
         gsettings.set(gs_schema, gs_key, gs_value)
 
+    def disable_screen_blanking(self):
+        gs_schema = 'org.gnome.desktop.session'
+        gs_key = 'idle-delay'
+        gs_previous = '%s/%s' % (gs_schema, gs_key)
+        if gs_previous in self.gsettings_previous:
+            return
+
+        gs_value = gsettings.get(gs_schema, gs_key)
+        self.gsettings_previous[gs_previous] = gs_value
+
+        if gs_value:
+            gsettings.set(gs_schema, gs_key, 0)
+
+        atexit.register(self.enable_screen_blanking)
+
+    def enable_screen_blanking(self):
+        gs_schema = 'org.gnome.desktop.session'
+        gs_key = 'idle-delay'
+        gs_previous = '%s/%s' % (gs_schema, gs_key)
+        gs_value = self.gsettings_previous[gs_previous]
+
+        gsettings.set(gs_schema, gs_key, gs_value)
+
     def disable_powermgr(self):
         gs_schema = 'org.gnome.settings-daemon.plugins.power'
         gs_key = 'active'
@@ -738,6 +830,7 @@ class Wizard(BaseFrontend):
 
         self.disable_volume_manager()
         self.disable_screensaver()
+        self.disable_screen_blanking()
         self.disable_powermgr()
 
         if 'UBIQUITY_ONLY' in os.environ:
@@ -1282,13 +1375,19 @@ class Wizard(BaseFrontend):
     def page_name(self, step_index):
         return self.steps.get_nth_page(step_index).get_name()
 
-    def toggle_next_button(self, label='gtk-go-forward'):
+    def toggle_next_button(self, label='gtk-go-forward', suggested=False):
         if label != 'gtk-go-forward':
             self.next.set_label(self.get_string(label))
         else:
             self.next.set_label(label)
             self.translate_widget(self.next)
 
+        style_context = self.next.get_style_context()
+        if suggested:
+            style_context.add_class('suggested-action')
+        else:
+            style_context.remove_class('suggested-action')
+
     def toggle_skip_button(self, label='skip'):
         self.skip.set_label(self.get_string(label))
         self.skip.show()
@@ -1331,11 +1430,15 @@ class Wizard(BaseFrontend):
                         hasattr(page.ui, 'plugin_on_skip_clicked'))
                     cur.show()
                     is_install = page.ui.get('plugin_is_install')
+                    is_restart = \
+                        page.ui.get('plugin_is_restart')
                     break
         if not cur:
             return False
 
-        if is_install and not self.oem_user_config:
+        if is_restart and not self.oem_user_config:
+            self.toggle_next_button('restart_button', suggested=True)
+        elif is_install and not self.oem_user_config:
             self.toggle_next_button('install_button')
         else:
             self.toggle_next_button()
@@ -1351,7 +1454,8 @@ class Wizard(BaseFrontend):
 
         if self.pagesindex == 0:
             self.allow_go_backward(False)
-        elif self.pages[self.pagesindex - 1].module.NAME == 'partman':
+        elif 'partman' in [page.module.NAME for page in
+                           self.pages[:self.pagesindex - 1]]:
             # We're past partitioning.  Unless the install fails, there is no
             # going back.
             self.allow_go_backward(False)
@@ -1643,6 +1747,17 @@ class Wizard(BaseFrontend):
                 (self.partitioned, self.timezone_set))
             return
 
+        # Setup zfs layout
+        use_zfs = self.db.get('ubiquity/use_zfs')
+        if use_zfs == 'true':
+            env = os.environ.copy()
+            zfs_keystore_key = self.db.get('ubiquity/zfs_keystore_key')
+            if zfs_keystore_key:
+                os.environ['ZFS_KS_KEY'] = zfs_keystore_key
+            misc.execute_root('/usr/share/ubiquity/zsys-setup', 'init')
+            os.environ.clear()
+            os.environ.update(env)
+
         syslog.syslog('Starting the installation')
 
         from ubiquity.debconfcommunicator import DebconfCommunicator
@@ -1776,14 +1891,22 @@ class Wizard(BaseFrontend):
             msg = title
         dialog = Gtk.MessageDialog(
             self.live_installer, Gtk.DialogFlags.MODAL,
-            Gtk.MessageType.ERROR, Gtk.ButtonsType.OK, msg)
+            Gtk.MessageType.ERROR, Gtk.ButtonsType.OK, "")
         dialog.set_title(title)
+        dialog.set_markup(msg)
+        for label in dialog.get_message_area().get_children():
+            if not isinstance(label, Gtk.Label):
+                continue
+            label.connect('activate-link', self.on_link_clicked)
         dialog.run()
         self.set_busy_cursor(saved_busy_cursor)
         dialog.hide()
         if fatal:
             self.return_to_partitioning()
 
+    def on_link_clicked(self, widget, uri):
+        misc.launch_uri(uri)
+
     def toggle_grub_fail(self, unused_widget):
         if self.grub_no_new_device.get_active():
             self.no_grub_warn.show()
